#ifndef ONEWAYLIST_H
#define ONEWAYLIST_H

#ifdef UNIQUE
#include <memory>
#endif
#include <stdexcept>

template<typename T>
class Node {
public:
    T data;
#ifdef UNIQUE
    std::unique_ptr<Node<T>>
#else
    Node<T>*
#endif
        next;
    Node();
    Node(T value);
};

template<typename T>
class OnewayList {
#ifdef UNIQUE
    std::unique_ptr<Node<T>>
#else
    Node<T>*
#endif
        head;
public:
    OnewayList();
    OnewayList(T);
    void append(T);
    bool remove(int, bool = 0);
    T get(int) const;
    ~OnewayList();
    int find(T, bool = 0) const;
    T& operator[](int) const;
};

template<typename T>
Node<T>::Node() : data(T()), next(nullptr) {}

template<typename T>
Node<T>::Node(T value) : data(value), next(nullptr) {}

template<typename T>
OnewayList<T>::OnewayList() : head(nullptr) {}

template<typename T>
OnewayList<T>::OnewayList(T value) {
#ifdef UNIQUE
    head = std::make_unique<Node<T>>(value);
#else
    head = new Node<T>(value);
#endif
}

template<typename T>
void OnewayList<T>::append(T value) {
    if (head == nullptr) {
#ifdef UNIQUE
        head = std::make_unique<Node<T>>(value);
#else
        head = new Node<T>(value);
#endif
        return;
    }
    Node<T>* temp = head
#ifdef UNIQUE
                        .get()
#endif
        ;
    while (temp->next != nullptr) {
        temp = temp->next
#ifdef UNIQUE
                   .get()
#endif
            ;
    }
    temp->next =
#ifdef UNIQUE
        std::make_unique<Node<T>>(value);
#else
        new Node<T>(value);
#endif
}

template<typename T>
bool OnewayList<T>::remove(int pos, bool soft) {
    if (!head) {
        if (soft) return false;
        else throw std::runtime_error("Trying to remove element in an empty list");
    }
    
    if (pos == 0) {
        Node<T>* temp = head
#ifdef UNIQUE
                            .get()
#endif
            ;
        if (head->next) head =
#ifdef UNIQUE
                std::move(head->next);
#else
                head->next;
#endif
        else head = nullptr;
#ifndef UNIQUE
        delete temp;
#endif
        return true;
    }
    Node<T>* prev;
    Node<T>* current = head
#ifdef UNIQUE
                           .get()
#endif
        ;
    for (int i = 0; i < pos; i++) {
        if (current == nullptr) throw std::out_of_range("Trying to remove non-existing element");
        prev = current;
        current = current->next
#ifdef UNIQUE
                      .get()
#endif
            ;
    }
    
    if (current->next) prev->next =
#ifdef UNIQUE
            std::move(current->next);
#else
            current->next;
    delete current;
#endif
    return true;
}

template<typename T>
T OnewayList<T>::get(int pos) const {
    Node<T>* temp = head
#ifdef UNIQUE
                        .get()
#endif
        ;
    int index = 0;
    while (temp != nullptr && index != pos) {
        temp = temp->next
#ifdef UNIQUE
                   .get()
#endif
            ;
        index++;
    }
    
    if (temp == nullptr) throw std::runtime_error("Trying to access non-existing key");
    return temp->data;
}

template<typename T>
OnewayList<T>::~OnewayList() {
    while (this->remove(0, true));
}

template<typename T>
int OnewayList<T>::find(T findme, bool soft) const {
    if (!head) {
        if (soft) return -1;
        else throw std::runtime_error("Trying to find entry in empty list");
    }
    Node<T>* curp = head
#ifdef UNIQUE
                        .get()
#endif
        ;
    int pos = 0;
    while (curp->data != findme) {
        curp = curp->next
#ifdef UNIQUE
                   .get()
#endif
            ;
        if (curp) pos++;
        else if (soft) return -1;
        else throw std::runtime_error(
                #ifndef REV
                "No entry " + std::to_string(findme) + " found in list"
                #else
                "No entry found in list"
                #endif
                );
    }
    return pos;
}
template<typename T>
T& OnewayList<T>::operator[](int pos)const {
    Node<T>* temp = head
#ifdef UNIQUE
                        .get()
#endif
        ;
    int index = 0;
    while (temp != nullptr && index != pos) {
        temp = temp->next
#ifdef UNIQUE
                   .get()
#endif
            ;
        index++;
    }
    
    if (temp == nullptr) throw std::runtime_error("Trying to access non-existing key with operator[]");
    return temp->data;
}
#endif // ONEWAYLIST_H
